// +------------------------------------------------------------------------------------------+
// |  ARC_API-3.6.2-Example.cpp                                                               |
// +------------------------------------------------------------------------------------------+
// |  The path to the CArcBase and CArcDevice librarys must be set before this code can be    |
// |  executed:                                                                               |
// |                                                                                          |
// |  Windows:                                                                                |
// |  SET PATH=C:\SomePath\ARC_API\3.6.2                                                      |
// |                                                                                          |
// |  Linux:                                                                                  |
// |  SET PATH=/SomePath/Projects/ARC_API/3.6.2                                               |
// +------------------------------------------------------------------------------------------+
#include <iostream>
#include <stdexcept>
#include <filesystem>
#include <memory>
#include <cstdlib>
#include <string>

#include <CArcDevice.h>
#include <CArcPCIe.h>
#include <ArcDefs.h>


// +-----------------------------------------------------------------------------+
// |  Comment out to use the setupController() method instead of performing a    |
// |  manual setup.                                                              |
// +-----------------------------------------------------------------------------+
//#define ARC_MANUAL_SETUP


// +-----------------------------------------------------------------------------+
// |  This class handles the callbacks for an exposure                           |
// +-----------------------------------------------------------------------------+
class CExposeListener : public arc::gen3::CExpIFace
{
	public:

		CExposeListener( void ) = default;

		void exposeCallback( [[maybe_unused]] float fElapsedTime )
		{
		}

		virtual void readCallback( [[maybe_unused]] std::uint32_t uiPixelCount )
		{
			std::cout << "pixel count: " << uiPixelCount << "\n";
		}
};



// +-----------------------------------------------------------------------------+
// |  Main function. Initializes a controller and takes an exposure.             |
// |                                                                             |
// |  Usage:  ARC_API-3.6.2-Example <timing file>                                |
// +-----------------------------------------------------------------------------+
int main( int iArgc, char** pszArgs )
{
	//
	// Process command line
	//
	if ( iArgc != 2 )
	{
		std::cout << "\nUsage:  ARC_API-3.6.2-Example <timing file>\n";

		return EXIT_SUCCESS;
	}

	std::filesystem::path tTimFile = pszArgs[ 1 ];

	try
	{
		if ( !std::filesystem::exists( tTimFile ) || !std::filesystem::is_regular_file( tTimFile ) )
		{
			std::cerr << "\nERROR: Invalid timing file!\n";

			return EXIT_SUCCESS;
		}
	}
	catch ( const std::filesystem::filesystem_error& fse )
	{
		std::cerr << "\nERROR: Invalid timing file! " << fse.what() << '\n';

		return EXIT_SUCCESS;
	}

	//
	// Look for ARC-66 PCIe devices
	//
	try
	{
		arc::gen3::CArcPCIe::findDevices();

		std::cout << "[] Found ARC66 PCIe: " << arc::gen3::CArcPCIe::deviceCount() << '\n';
	}
	catch ( ... )
	{
		std::cout << "[] No Device Found! :(\n";

		return EXIT_SUCCESS;
	}

	//
	// Create a CArcDevice object and wrap into a smart pointer so you don't have to call delete.
	//
	std::unique_ptr<arc::gen3::CArcDevice> pArcDev( new arc::gen3::CArcPCIe() );

	constexpr auto uiRows = 1024;
	constexpr auto uiCols = 1200;

	std::cout << "[] Opening device ... ";
	pArcDev->open( 0, uiCols * uiCols * 2 );
	std::cout << "done!\n";

	if ( pArcDev->isControllerConnected() )
	{
		std::cout << "[] Yeah! A controller is connected!\n";
	}

	else
	{
		std::cout << "[] Hmmm, maybe we forgot to turn it on!\n";
	}

	try
	{
		// NOTE: Zero is correct for GenIII. This is NOT the same id as TIM_ID. It's
		//       used to distinquish the type of controller.  You shouldn't use this.
		std::cout << "[] Controller ID: " << pArcDev->getControllerId() << '\n';

		// This is correct ... in ASCII it's 'ARC6'
		std::cout << "[] Board ID: " << std::hex << pArcDev->getId() << std::dec << '\n';

#ifdef ARC_MANUAL_SETUP

		auto uiReply = 0U;

		std::cout << "[] Resetting controller ... ";
		pArcDev->resetController();
		std::cout << "done!\n";

		// Send a series of Test Data Links ( TDL ) to timing board.
		for ( auto i = 0U; i < 123; i++ )
		{
			auto uiRetVal = pArcDev->command( { arc::TIM_ID, arc::TDL, i } );

			if ( uiRetVal != i ) { throw std::runtime_error( "TDL failed!\n" ); }
		}
		std::cout << "[] TDL succeed!\n";

		// Windows requires a "\\", not a "\" for filenames. So you should use "C:\\tim.lod"
		std::cout << "[] Loading controller file ... ";
		pArcDev->loadControllerFile( tTimFile.string() );
		std::cout << "done!\n";

		// Send Power On to the controller
		std::cout << "[] Powering controller ... ";
		if ( ( uiReply = pArcDev->command( { arc::TIM_ID, arc::PON } ) ) != arc::DON )
		{
			throw std::runtime_error( "PON failed!" );
		}
		std::cout << "done!\n";

#else

		std::cout << "[] Performing controller setup ... ";
		pArcDev->setupController( true, true, true, uiRows, uiCols, tTimFile.string() );
		std::cout << "done!\n";

#endif

		std::cout << "[] Starting exposure ... \n";
		pArcDev->expose( 0.f, uiRows, uiCols, nullptr, new CExposeListener() );
	}
	catch ( const std::exception& exc )
	{
		std::cerr << '\n' << exc.what() << '\n';
	}

	return 0;
}
